/*
 * File: killring.c
 * Date: 2011-11-20
 *
 * Evsiz (evsiz@this-voice.org)
 * Rho released under the Creative Commons NonCommercial ShareAlike License 3.0
 * See http://creativecommons.org for details
 *
 * Rho is a fork of the Linux port of MicroEMACS developed by Humberto
 * Ortiz-Zuazaga and Jean-Pierre Demailly. Their work forked the MicroEMACS
 * of Daniel M. Lawrence which was in turn a fork of uemacs by Dave G.
 * Conroy. MicroEmacs was copyrighted 1993 by Mr. Lawrence with a
 * stipulation that it could not be used commercially. Uemacs was released
 * by Mr. Conroy into the public domain in 1985.
 *
 */


/***

The kill ring is a linked-list of buffers holding deleted text.
Or it was supposed to be.  The ring of buffers never happened,
so far as I can tell.  Just one kill buffer is all I see here.
All the multiple buffer stuff was ifdef'd off or not connected
to anything.  These are the working bits.

***/

#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	"rho.h"
#include	"killring.h"
#include	"cursor.h"
#include	"line.h"
#include	"messages.h"
#include	"region.h"
#include	"utils.h"




static long last_size = -1L;    /* last # of bytes yanked */

/*++

Delete all text in a kill buffer. Called by commands when a new kill
context is being created. The kill buffer array is released, just in
case the buffer has grown to immense size.

++*/
static void kdelete()
{
    KILL *kp;                   /* ptr to scan kill buffer chunk list */

    if (kbufh[kill_index] != NULL) {

        /* first, delete all the chunks */
        kbufp[kill_index] = kbufh[kill_index];
        while (kbufp[kill_index] != NULL) {
            kp = kbufp[kill_index]->d_next;
            free((char *) kbufp[kill_index]);
            kbufp[kill_index] = kp;
        }

        /* and reset all the kill buffer pointers */
        kbufh[kill_index] = kbufp[kill_index] = NULL;
        kskip[kill_index] = 0;
        kused[kill_index] = KBLOCK;
    }
}



/*++

Advance to the next position in the kill ring, pushing the current kill
buffer and clearing what will be the new kill buffer

++*/
void next_kill()
{
    /* advance to the next kill ring entry */
    kill_index++;
    if (kill_index == NRING)
        kill_index = 0;

    /* and clear it, so it is ready for use */
    kdelete();
}



/*++

Insert a character to the kill buffer, allocating new chunks as needed.
Return TRUE if all is well, and FALSE on errors.

++*/
int kinsert(int direct, char c)
{
    KILL *nchunk;               /* ptr to newly malloced chunk */

    if (direct == FORWARD) {

        /* check to see if we need a new chunk */
        if (kused[kill_index] >= KBLOCK) {
            nchunk = (KILL *) malloc(sizeof(KILL));

            if (kbufh[kill_index] == NULL)      /* set head ptr if first time */
                kbufh[kill_index] = nchunk;
            if (kbufp[kill_index] != NULL)      /* point the current to this new one */
                kbufp[kill_index]->d_next = nchunk;
            kbufp[kill_index] = nchunk;
            kbufp[kill_index]->d_next = NULL;
            kused[kill_index] = 0;
        }

        /* and now insert the character */
        kbufp[kill_index]->d_chunk[kused[kill_index]++] = c;
    } else {
        /* REVERSE */
        /* check to see if we need a new chunk */
        if (kskip[kill_index] == 0) {
            nchunk = (KILL *) malloc(sizeof(KILL));

            if (kbufh[kill_index] == NULL) {    /* set head ptr if first time */
                kbufh[kill_index] = nchunk;
                kbufp[kill_index] = nchunk;
                kskip[kill_index] = KBLOCK;
                kused[kill_index] = KBLOCK;
                nchunk->d_next = (KILL *) NULL;
            } else {
                nchunk->d_next = kbufh[kill_index];
                kbufh[kill_index] = nchunk;
                kskip[kill_index] = KBLOCK;
            }
        }

        /* and now insert the character */
        kbufh[kill_index]->d_chunk[--kskip[kill_index]] = c;
    }
    return (TRUE);
}


#define	Char_insert(a)	(a == '\r' ? insertnewline() : insertnchars(1, a))

/*++

Yank text back from the kill buffer. This is really easy. All of the work
is done by the standard insert routines. All you do is run the loop, and
check for errors. Bound to "C-Y".

++*/
int yank(f, n)
int f, n;

{

    char *sp;                   /* pointer into string to insert */
    int counter;                /* counter into kill buffer data */
    int curoff;                 /* storage for line before yanking */
    KILL *kptr;                 /* pointer into kill buffer */
    LINE *curline;

    f = 0;                      /* splint */

    if (curbp->b_mode & MDVIEW) /* don't allow this command if  */
        return (err_readonly());        /* we are in read only mode     */

    if (n < 0)
        return (FALSE);

    /* make sure there is something to yank */
    if (kbufh[kill_index] == NULL) {
        last_size = 0L;
        return (TRUE);          /* not an error, just nothing */
    }

    /*
     * Save the local pointers to hold global ".".
     */
    if (yankflag) {
        /* Find the *previous* line, since the line we are on
         * may disappear due to re-allocation.  This works even
         * if we are on the first line of the file.
         */
        curline = lback(curwp->w_dotp);
        curoff = curwp->w_doto;
    }

    /* for each time.... */
    while (n--) {
        last_size = 0L;
        if (kskip[kill_index] > 0) {
            kptr = kbufh[kill_index];
            sp = &(kptr->d_chunk[kskip[kill_index]]);
            counter = kskip[kill_index];
            while (counter++ < KBLOCK) {
                Char_insert(*sp);
                last_size++;
                ++sp;
            }
            kptr = kptr->d_next;
        } else {
            kptr = kbufh[kill_index];
        }

        if (kptr != (KILL *) NULL) {
            while (kptr != kbufp[kill_index]) {
                sp = kptr->d_chunk;
                for (counter = 0; counter < KBLOCK; counter++) {
                    Char_insert(*sp);
                    last_size++;
                    ++sp;
                }
                kptr = kptr->d_next;
            }
            counter = kused[kill_index];
            sp = kptr->d_chunk;
            while (counter--) {
                Char_insert(*sp);
                last_size++;
                ++sp;
            }
        }
    }

    /* If requested, set global "." back to the
     * beginning of the yanked text.
     */
    if (yankflag) {
        curwp->w_dotp = lforw(curline);
        curwp->w_doto = curoff;
    }
    thisflag |= CFYANK;
    return (TRUE);
}



/*
 * Kill text. If called without an argument, it kills from dot to the end of
 * the line, unless it is at the end of the line, when it kills the newline.
 * If called with an argument of 0, it kills from the start of the line to dot.
 * If called with a positive argument, it kills from dot forward over that
 * number of newlines. If called with a negative argument it kills backwards
 * that number of newlines. Normally bound to "C-K".
 */

int killtext(f, n)
int f, n;
{
    LINE *nextp;
    long chunk;

    if (curbp->b_mode & MDVIEW) /* don't allow this command if  */
        return (err_readonly());        /* we are in read only mode     */
    if ((lastflag & CFKILL) == 0)       /* Clear kill buffer if */
        next_kill();            /* last wasn't a kill.  */
    thisflag |= CFKILL;

    if (f == FALSE) {
        chunk = lused(curwp->w_dotp) - curwp->w_doto;
        if (chunk == 0)
            chunk = 1;
    } else if (n == 0) {
        chunk = -curwp->w_doto;
    } else if (n > 0) {
        chunk = lused(curwp->w_dotp) - curwp->w_doto + 1;
        nextp = lforw(curwp->w_dotp);
        while (--n) {
            if (nextp == curbp->b_linep)
                return (FALSE);
            chunk += lused(nextp) + 1;
            nextp = lforw(nextp);
        }
    } else if (n < 0) {
        chunk = -curwp->w_doto;
        nextp = lback(curwp->w_dotp);
        while (n++) {
            if (nextp == curbp->b_linep)
                return (FALSE);
            chunk -= lused(nextp) + 1;
            nextp = lback(nextp);
        }
    }
    return (deletenchars(chunk, TRUE));
}


/*++

Delete n paragraphs starting with the current one

++*/
int kill_paragraph(f, n)
int f;
int n;                          /* # of paras to delete */

{
    int status;                 /* returned status of functions */

    f = 0;                      /* splint */

    while (n--) {               /* for each paragraph to delete */

        /* mark out the end and beginning of the para to delete */
        next_paragraph(FALSE, 1);

        /* set the mark here */
        curwp->w_markp[0] = curwp->w_dotp;
        curwp->w_marko[0] = curwp->w_doto;

        /* go to the beginning of the paragraph */
        previous_paragraph(FALSE, 1);
        curwp->w_doto = 0;      /* force us to the beginning of line */

        /* and delete it */
        if ((status = kill_region(FALSE, 1)) != TRUE)
            return (status);

        /* and clean up the 2 extra lines */
        deletenchars(2L, TRUE);
    }
    return (TRUE);
}
