/*
 * File: ufunc.c
 * Date: 2011-11-01
 *
 * Evsiz (evsiz@this-voice.org)
 * Rho released under the Creative Commons NonCommercial ShareAlike License 3.0
 * See http://creativecommons.org for details
 *
 * Rho is a fork of the Linux port of MicroEMACS developed by Humberto
 * Ortiz-Zuazaga and Jean-Pierre Demailly. Their work forked the MicroEMACS
 * of Daniel M. Lawrence which was in turn a fork of uemacs by Dave G.
 * Conroy. MicroEmacs was copyrighted 1993 by Mr. Lawrence with a
 * stipulation that it could not be used commercially. Uemacs was released
 * by Mr. Conroy into the public domain in 1985.
 *
 */

/***

The original MicroEMACS code exploits string arrays and #defines to create
primitive dictionary structures.  This file uses ufuncs[] in this way.
The array ufuncs is initialized here and enumerated with defines for
other files in rho.h.  This file has sections for both UFUNCS and the
user vars UVARS which are the vars users can create in macros.

***/


#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	"rho.h"
#include	"ufuncs.h"
#include	"bind.h"
#include	"envars.h"
#include	"exec.h"
#include	"fileio.h"
#include	"macros.h"
#include	"messages.h"
#include	"utils.h"




/******************************* UFUNCS **********************************/

/*
 *
 * Each line here loads a UFUNC, which has a char* naming the ufunc and
 * a constant for how many params are in the ufunc.  Unlike the KEYTAB's
 * EPOINTERS which were a hack, this seems a perfectly good way to load
 * the UFUNCS.  So we'll tell splint to ignore it.
 *
 */

/*@ignore@*/

UFUNC ufuncs[] = {
    {"abs", MONAMIC}
    ,                           /* absolute value of a number */
    {"add", DYNAMIC}
    ,                           /* add two numbers together */
    {"and", DYNAMIC}
    ,                           /* logical and */
    {"bin", MONAMIC}
    ,                           /* lookup what function name is bound to a key */
    {"cat", DYNAMIC}
    ,                           /* concatenate string */
    {"chr", MONAMIC}
    ,                           /* integer to char conversion */
    {"div", DYNAMIC}
    ,                           /* division */
    {"env", MONAMIC}
    ,                           /* retrieve a system environment var */
    {"equ", DYNAMIC}
    ,                           /* logical equality check */
    {"exi", MONAMIC}
    ,                           /* check if a file exists */
    {"fin", MONAMIC}
    ,                           /* look for a file on the PATH */
    {"gre", DYNAMIC}
    ,                           /* logical greater than */
    {"gtk", NILNAMIC}
    ,                           /* get one character */
    {"ind", MONAMIC}
    ,                           /* evaluate indirect value */
    {"lef", DYNAMIC}
    ,                           /* left string(string, len) */
    {"les", DYNAMIC}
    ,                           /* logical less than */
    {"low", MONAMIC}
    ,                           /* lower case string */
    {"mid", TRINAMIC}
    ,                           /* mid string(string, pos, len) */
    {"mod", DYNAMIC}
    ,                           /* mod */
    {"neg", MONAMIC}
    ,                           /* negate */
    {"not", MONAMIC}
    ,                           /* logical not */
    {"or", DYNAMIC}
    ,                           /* logical or */
    {"rig", DYNAMIC}
    ,                           /* right string(string, pos) */
    {"seq", DYNAMIC}
    ,                           /* string logical equality check */
    {"sgr", DYNAMIC}
    ,                           /* string logical greater than */
    {"sle", DYNAMIC}
    ,                           /* string logical less than */
    {"sub", DYNAMIC}
    ,                           /* subtraction */
    {"tim", DYNAMIC}
    ,                           /* multiplication */
    {"upp", MONAMIC}
    ,                           /* uppercase string */
};

/*@end@*/


/*++

UFUNC return absolute value of int. Also local helper.

++*/
static int abs_val(x)
int x;

{
    return (x < 0 ? -x : x);
}



/*++

Return f_name string of ufunc in bind_ufunc.h

++*/
static char *get_ufuncs_fname(i)
int i;

{
    return (ufuncs[i].f_name);
}



/*++

Match a string to a "name" in names[] (rho_name.h)

++*/
static char *str2fn_str(skey)
char *skey;

{
    char *fn_string;

    fn_string = get_name(exist_binding(keystr2keycmd(skey)));
    if (fn_string == NULL)
        fn_string = (char *) errorm;

    return (fn_string);
}



/*++

Evaluate an element of UFUNC ufuncs[] (bind_ufunc.h)

++*/
char *eval_ufunc(fname)
char *fname;                    /* name of function to evaluate */

{
    int fnum;                   /* index to function to eval */
    int arg;                    /* value of some arguments */
    char arg1[NSTRING];         /* value of first argument */
    char arg2[NSTRING];         /* value of second argument */
    char arg3[NSTRING];         /* value of third argument */
    static char result[2 * NSTRING];    /* string result */

    /* look the function up in the function table */
    fname[3] = 0;               /* only first 3 chars significant */
    (void) mklower(fname);      /* and let it be upper or lower case */
    fnum = index_of_str(fname, get_ufuncs_fname, NUFUNCS);

    /* return errorm on a bad reference */
    if (fnum == -1) {
        mlwrite(TEXT244, fname);
        /* "%%No such function as '%s'" */
        return ((char *) errorm);
    }

    /* if needed, retrieve the first argument */
    if (ufuncs[fnum].f_type >= MONAMIC) {
        if (get_macro_line(arg1) != TRUE)
            return ((char *) errorm);

        /* if needed, retrieve the second argument */
        if (ufuncs[fnum].f_type >= DYNAMIC) {
            if (get_macro_line(arg2) != TRUE)
                return ((char *) errorm);

            /* if needed, retrieve the third argument */
            if (ufuncs[fnum].f_type >= TRINAMIC)
                if (get_macro_line(arg3) != TRUE)
                    return ((char *) errorm);
        }
    }


    switch (fnum) {
    case UFABS:
        return (int2str(abs_val(str2int(arg1))));
    case UFADD:
        return (int2str(str2int(arg1) + str2int(arg2)));
    case UFAND:
        return (bool2str(str2bool(arg1) && str2bool(arg2)));
    case UFBIND:
        return (str2fn_str(arg1));
    case UFCAT:
        strcpy(result, arg1);
        return (strcat(result, arg2));
    case UFCHR:
        result[0] = str2int(arg1);
        result[1] = 0;
        return (result);
    case UFDIV:
        if ((arg = str2int(arg2)) != 0)
            return (int2str(str2int(arg1) / arg));
        else {
            mlwrite(TEXT245);
            /* "%%Division by Zero is illegal" */
            return ((char *) errorm);
        }
    case UFENV:
        return (fixnull(getenv(arg1)));
    case UFEQUAL:
        return (bool2str(str2int(arg1) == str2int(arg2)));
    case UFEXIST:
        return (bool2str(fexist(arg1)));
    case UFFIND:
        return (fixnull(search_paths(arg1, TRUE)));
    case UFGREATER:
        return (bool2str(str2int(arg1) > str2int(arg2)));
    case UFGTKEY:
        result[0] = tgetc();
        result[1] = 0;
        return (result);
    case UFIND:
        return (strcpy(result, fixnull(get_TKval(arg1))));
    case UFLEFT:
        return (strncpy(result, arg1, (size_t) str2int(arg2)));
    case UFLESS:
        return (bool2str(str2int(arg1) < str2int(arg2)));
    case UFLOWER:
        return (mklower(arg1));
    case UFMID:
        arg = str2int(arg2);
        if (arg > (int) strlen(arg1))
            arg = (int) strlen(arg1);
        return (strncpy(result, &arg1[arg - 1], (size_t) str2int(arg3)));
    case UFMOD:
        if ((arg = str2int(arg2)) != 0)
            return (int2str(str2int(arg1) % arg));
        else {
            mlwrite(TEXT245);
            /* "%%Division by Zero is illegal" */
            return ((char *) errorm);
        }
    case UFNEG:
        return (int2str(-str2int(arg1)));
    case UFNOT:
        return (bool2str(str2bool(arg1) == FALSE));
    case UFOR:
        return (bool2str(str2bool(arg1) || str2bool(arg2)));
    case UFRIGHT:
        arg = str2int(arg2);
        if (arg > (int) strlen(arg1))
            arg = (int) strlen(arg1);
        return (strcpy(result, &arg1[strlen(arg1) - arg]));
    case UFSEQUAL:
        return (bool2str(strcmp(arg1, arg2) == 0));
    case UFSGREAT:
        return (bool2str(strcmp(arg1, arg2) > 0));
    case UFSLESS:
        return (bool2str(strcmp(arg1, arg2) < 0));
    case UFSUB:
        return (int2str(str2int(arg1) - str2int(arg2)));
    case UFTIMES:
        return (int2str(str2int(arg1) * str2int(arg2)));
    case UFUPPER:
        return (mkupper(arg1));
    }
    return ("");                /* should not get here */
}



/******************************* UVARS  **********************************/



/*++

Initialize the user variable list.

++*/
void varinit()
{
    int i;

    for (i = 0; i < MAXUVARS; i++)
        uv[i].u_name[0] = 0;
}



/*++

Look up a user var's value

++*/
char *get_uvar(vname)
char *vname;                    /* name of user variable to fetch */

{
    int vnum;                   /* ordinal number of user var */
    char *vptr;                 /* temp pointer to function value */

    /* limit comparisons to significant length */
    if (strlen(vname) >= NVSIZE)        /* "%" counts, but is not passed */
        vname[NVSIZE - 1] = '\0';

    /* scan the list looking for the user var name */
    for (vnum = 0; vnum < MAXUVARS; vnum++) {
        if (uv[vnum].u_name[0] == 0)
            return ((char *) errorm);
        if (strcmp(vname, uv[vnum].u_name) == 0) {
            vptr = uv[vnum].u_value;
            if (vptr)
                return (vptr);
            else
                return ((char *) errorm);
        }
    }

    /* return errorm if we run off the end */
    return ((char *) errorm);
}
